
LIMIT := 16;

##
## F = GF(2)
## m and l are monomials (written as lists [i1, ..., ik])
## v and w are polynomials (written as lists of monomials)
##

lt := function( m, l )
    local dm, dl;
    if m = l then return false; fi;
    dm := Length(m);
    dl := Length(l);
    if dm <> dl then return dm < dl; fi;
    return m > l;
end;

le := function( m, l )
    if m=l then return true; fi;
    return lt(m,l);
end;

##
## find maximal monomial
##
FindMax := function( v )
    local m, i;
    m := v[1];
    for i in [2..Length(v)] do
        if IsBound(v[i]) and lt(m, v[i]) then m := v[i]; fi;
    od;
    return m;
end;

##
## find minimal monomial
##
FindMin := function( v )
    local m, i;
    m := v[1];
    for i in [2..Length(v)] do
        if IsBound(v[i]) and lt(v[i], m) then m := v[i]; fi;
    od;
    return m;
end;

##
## determine normal form; that is, write the polynomial v as compacted
## and sorted list without duplicates
##
NormalForm := function(v)
    local w, k, i, j;
    w := [];
    k := 1;
    for i in [1..Length(v)] do
        j := Position(w, v[i]);
        if IsBool(j) then
            w[k] := v[i];
            k := k+1;
        else
            Unbind(w[j]);
        fi;
    od;
    w := Compacted(w);
    Sort(w, lt);
    return w;
end;

##
## translate from ordinary polynomials to my polynomials
##
## this function evaluates x^p = x and also orders the monomials 
## with respect to lt.
##
ExtExtRep := function( poly )
    local e, v, j, i;
    e := ExtRepPolynomialRatFun(poly);
    v := [];
    for i in [1,3..Length(e)-1] do
        v[(i+1)/2] := e[i]{[1,3..Length(e[i])-1]};
    od;
    return NormalForm(v);
end;

##
## add or subtract two polynomials
##
AddOrSubtract := function(v, w)
    return NormalForm( Concatenation(v,w) );
end;

##
## Leading term of a polynomial; if flag = true, then the polynomial
## is in normal form.
##
LeadingTerm := function(v, flag)
    if flag then return v[Length(v)]; fi;
    return FindMax(v);
end;

##
## given a list of polynomials in normalform and a monomial m, find
## the nicest polynomial with leading term m
##
Pivot := function( vs, m )
    local j, i;
    j := 0;
    for i in [1..Length(vs)] do
        if LeadingTerm(vs[i], true) = m then
            if j = 0 or Length(vs[i]) < Length(vs[j]) then
                j := i;
            fi;
        fi;
    od;
    return j;
end;

##
## apply Gaussian elimination to a list of polynomials; assume that
## the polynomials are in normal form
##
Linearise := function( vs )
    local done, exts, m, l, j, i;
    done := [];
    exts := Filtered(vs, x -> Length(x)>0);
    while Length(exts) > 0 do
        l := List( exts, x -> LeadingTerm(x,true));
        m := FindMax(l);
        #Print("consider ",m," with ",Length(exts)," and ",Length(done),"\n");
        j := Pivot(exts, m);
        Add(done, ShallowCopy(exts[j]));
        for i in [1..Length(exts)] do
            if i <> j and LeadingTerm(exts[i],true) = m then
                exts[i] := AddOrSubtract(exts[i], exts[j]);
            fi;
        od;
        exts[j] := [];
        exts := Filtered(exts, x -> Length(x)>0);
    od;

    for i in [1..Length(done)] do
        m := LeadingTerm(done[i], true);
        for j in [1..i-1] do
            if m in done[j] then
                done[j] := AddOrSubtract(done[j], done[i]);
            fi;
        od;
    od;

    return done;
end;


##
## Eliminate variable i in all polynomials in vs
##
Eliminate := function( vs, i, a )
    local j, ns;
    ns := 0*vs;
    if a = 0*a then
        for j in [1..Length(vs)] do
            ns[j] := Filtered(vs[j], x -> not i in x);
            ns[j] := NormalForm(ns[j]);
        od;
    else
        for j in [1..Length(vs)] do
            ns[j] := List(vs[j], x -> Difference(x, [i]));
            ns[j] := NormalForm(ns[j]);
        od;
    fi;
    return Filtered(ns, x -> Length(x) > 0);
end;

##
## Definitions: for each variable i try to find a polynomials v
## in vs that defines i; that is, that has the form v = i+w where
## w is independent of i.
##
FindNext := function( defs, rems, vs )
    local i, r, t, s, j, v;

    if Length(rems)=0 then return fail; fi;

    # first try simple case
    for i in [1..Length(rems)] do
        r := rems[i];
        s := Filtered([1..Length(defs)], x-> r in defs[x]);
        s := Filtered(s, x -> Length(defs[x]) = 1);
#Print("try ",r," simple \n");
        for j in s do
            v := vs[j];
            if Length(Intersection(Set(Flat(v)), rems)) = 1 then
                return [r, j];
            fi;
        od;
    od;


    # now try the complicated case
    for i in [1..Length(rems)] do
        r := rems[i];
        s := Filtered([1..Length(defs)], x-> r in defs[x]);
        s := Filtered(s, x -> Length(defs[x]) > 1);
#Print("try ",r," non-simple \n");
        for j in s do
            v := vs[j];
            if Length(Intersection(Set(Flat(v)), rems)) = 1 then
                return [r, j];
            fi;
        od;
    od;

    return fail;
end;

Definitions := function(vs)
    local defs, i, v, j, rems, news, f;

    # set up and get all kinds of possibilities
    defs := List(vs, x -> []);
    for i in [1..Length(vs)] do
        v := vs[i];
        for j in [1..Length(v)] do
            if Length(v[j])=1 and
               ForAll([1..j-1], x -> not v[j][1] in v[x]) and
               ForAll([j+1..Length(v)], x -> not v[j][1] in v[x])
             then
               Add(defs[i], v[j][1]);
             fi;
        od;
    od;

    # resolve
    rems := Set(Flat(defs));
    news := [];
    repeat
        f := FindNext( defs, rems, vs );
        if not IsBool(f) then
            Add( news, [f[1], AddOrSubtract(vs[f[2]], [[f[1]]])] );
            defs[f[2]] := [];
            rems := Difference( rems, [f[1]] );
        else
            rems := rems{[1..Length(rems)-1]};
        fi;
    until Length(rems) = 0;


    # that's it
    return news;
end;

##
## Find singles; that is, polynomials of the form [[i]]
##
FindSingles := function( vs )
    local sing, v;
    sing := [];
    for v in vs do
        if Length(v) = 1 and Length(v[1])=1 then Add(sing, v[1][1]); fi;
    od;
    return sing;
end;

##
## Find singles; that is, polynomials of the form [[],[i]]
##
FindAntiSingles := function(vs)
    local sing, v;
    sing := [];
    for v in vs do
        if Length(v) = 2 then
            Sort(v);
            if v[1] = [] and Length(v[2]) = 1 then
                Add(sing, v[2][1]);
            fi;
        fi;
    od;
    return sing;
end;

##
## Evaluate polynomial v in vars f at e
##
Evaluate := function( v, f, e )
    local r, i, s, j, k;
    r := Zero(GF(2));
    for i in [1..Length(v)] do
        s := One(GF(2));
        for j in [1..Length(v[i])] do
            k := Position(f, v[i][j]);
            s := s*e[k];
        od;
        r := r+s;
    od;
    return r;
end;

IsCommonZero := function( vs, is, e )
    local v, f;
    for v in vs do
        f := Evaluate(v, is, e);
        if f <> 0*f then return false; fi;
    od;
    return true;
end;

Expand := function( e, is, rm, df )
    local f, i, j;
    f := List(is, x -> Zero(GF(2)));
    for i in [1..Length(rm)] do
        j := Position(is, rm[i]);
        f[j] := e[i];
    od;
    for i in [1..Length(df)] do
        j := Position(is, df[i][1]);
        f[j] := Evaluate( df[i][2], is, f );
    od;
    return f;
end;

VarsOfSystem := function(vs)
    return Set(Concatenation(vs));
end;

MatrixBySystem := function(vs)
    local vars, M, i, j, k;
    vars := VarsOfSystem(vs);
    Sort(vars, lt);
    M := NullMat(Length(vs), Length(vars), GF(2));
    for i in [1..Length(vs)] do
        for j in [1..Length(vs[i])] do
            k := Position(vars, vs[i][j]);
            M[i][k] := One(GF(2));
        od;
    od;
    return M;
end;


##
## Solve a system of polys
##
SolveSystem := function(vs)
    local is, sg, sa, df, rm, sl, ns, s1, s2, t1, t2, e, f, i, j, c, l;

    is := Set(Flat(vs));
    sa := [];

    # first linearise and eliminate singles
    repeat
        vs := Linearise(vs);
        sg := FindSingles(vs);
        if Length(sg) > 0 then
            for i in sg do
                vs := Eliminate(vs, i, Zero(GF(2)));
            od;
            Append(sa,sg);
        fi;
    until Length(sg) = 0;

    # now try to find definitions
    df := Definitions(vs);

    Print("-- solving system with ",Length(vs)," equations,");
    Print(Length(Set(Flat(vs))), " variables");
    Print(" and ",Length(df)," definitions \n");

    # check how many options are left
    rm := Difference( is, sa );
    rm := Difference( rm, List(df, x -> x[1]));

    if Length(rm) <= LIMIT then
        Print(" -- loop over remaining ",2^Length(rm)," options \n");
        sl := [];
        for e in GF(2)^Length(rm) do
            f := Expand( e, is, rm, df );
            if IsCommonZero( vs, is, f ) then Add(sl, f); fi;
        od;
        return sl;
    else
        c := Collected(Flat(vs));
        c := Filtered(c, x -> x[1] in rm);
        l := Maximum(List(c, x -> x[2]));
        c := List( Filtered(c, x -> x[2] = l), x -> x[1]);
        j := Random(c);
        Print(" -- choose ",j," in ", c,"\n");

        #Print("split up and recurse with ",rm[j]," = 0\n");
        ns := Eliminate(vs, j, 0);
        s1 := SolveSystem(ns);
        t1 := Set(Flat(ns));
        s1 := List( s1, x -> Expand( x, is, t1, [] ));

        #Print("split up and recurse with ",rm[j]," = 1\n");
        ns := Eliminate(vs, j, 1);
        s2 := SolveSystem(ns);
        t2 := Set(Flat(ns));
        s2 := List( s2, x -> Expand( x, is, t2, [[j,[[]]]] ));
        return Concatenation(s1, s2);
    fi;
end;


