
HasRepeatedFactor := function( mat )
    local f, c, X;
    X := Indeterminate(GF(2));
    f := CharacteristicPolynomial(mat);
    c := Collected(Factors(f));
    c := Filtered(c, x -> x[1] <> X);
    if Length(c) = 0 then return true; fi;
    return ForAny(c, x -> x[2] > 1);
end;

##
## the easy stuff
##
IsLieRelevant1 := function(mat)
    if TraceMat(mat) <> Zero(GF(2)) then return false; fi;
    if RankMat(mat) = Length(mat) then return false; fi;
    if not HasRepeatedFactor(mat) then return false; fi;
    return true;
end;

EigenvaluesMat := function( mat )
    local K, k, e, d, i;

    # determine Eigenspaces
    K := SplittingField(CharacteristicPolynomial(mat));
    k := Elements(K);
    e := List(k, x -> TriangulizedNullspaceMat(mat - x*mat^0));

    # enumerate
    d := [];
    for i in [1..Length(k)] do
        Append(d, List([1..Length(e[i])], x -> k[i]));
    od;
    return d;
end;

YieldsIdeal := function(dia, sum, c)
    local pos, i, j;
    pos := Filtered([1..Length(dia)], x -> dia[x] in c);
    for i in pos do
        for j in [1..Length(dia)] do
            if j<>i and (sum[i][j] <> fail) and (not sum[i][j] in pos) then
                return false;
            fi;
        od;
    od;
    return true;
end;

HasEigenIdealMat := function( mat )
    local dia, eig, sum, c, com, i;

    dia := EigenvaluesMat(mat);
    #if Length(dia)<Length(mat) then Error("should not happen"); fi;

    # check that there are interesting eigenvalues
    if ForAll(dia, x -> x in GF(2)) then return false; fi;

    # compute positions of sums
    sum := List(dia, x -> List(dia, y -> Position(dia,x+y)));

    # determine combinations of non-trivial eigenvalues
    eig := Set(Filtered(dia, x -> x <> Zero(GF(2))));
    for i in [1..Length(eig)] do
        com := Combinations(eig, i);
        for c in com do
            if YieldsIdeal(dia, sum, c) then return true; fi;
        od;
    od;
    return false;
end;

##
## use Eigenvalues of mat
##
IsLieRelevant2 := function(mat)
    return not HasEigenIdealMat(mat);
end;

Coeff := function(n, i, j)
    local c, l, k;
    c := 0;
    for l in [1..n] do
        for k in [l+1..n] do
            c := c + 1;
            if [l,k] = [i,j] then return c; fi;
        od;
    od;
end;

ExtAction := function( base, mat )
    local ind, dim, wed, c, i, j, k, d;

    # act on base
    ind := List(base, x -> SolutionMat(base, x*mat));

    # set up for wedge
    dim := Length(base)*(Length(base)-1)/2;
    wed := NullMat(dim,dim,GF(2));

    # construct wedge
    c := 0;
    for i in [1..Length(base)] do
        for j in [i+1..Length(base)] do
            c := c+1;
            for k in [1..Length(base)] do
                if j > k then 
                    d := Coeff(Length(base),k,j);
                    wed[c][d] := wed[c][d] + ind[i][k];
                elif j < k then 
                    d := Coeff(Length(base),j,k);
                    wed[c][d] := wed[c][d] + ind[i][k];
                fi;
                if i > k then 
                    d := Coeff(Length(base),k,i);
                    wed[c][d] := wed[c][d] + ind[j][k];
                elif i < k then 
                    d := Coeff(Length(base),i,k);
                    wed[c][d] := wed[c][d] + ind[j][k];
                fi;
            od;
        od;
    od;
    return wed;
end;

##
## use tensor and wedge action
##
IsLieRelevant3 := function(mat)
    local f, c, X, d, g, l, I, J, K, m, w, r, a, h;

    f := CharacteristicPolynomial(mat);
    c := Collected(Factors(f));
    X := Indeterminate(GF(2));
    
    # factor x
    d := Filtered(c, x -> x[1]=X);
    g := X^d[1][2];

    # factor x+1
    d := Filtered(c, x -> x[1]=X+X^0);
    if Length(d)>0 then
        h := (X+X^0)^d[1][2];
    else
        h := X^0;
    fi;
 
    if Degree(f) = Degree(g)+Degree(h) then 
        return Degree(h) + Degree(h)*(Degree(h)-1)/2 >= Length(mat);
    fi;

    # the rest
    l := f/(g*h);

    # get kernels 
    I := TriangulizedNullspaceMat(Value(l, mat));
    K := TriangulizedNullspaceMat(Value(h, mat));

    # determine wedge
    w := ExtAction(I, mat);
    r := CharacteristicPolynomial(w);
    
    # induce
    a := Value(r, mat);
    J := TriangulizedNullspaceMat(a);

    # check if L2 is an ideal
    if ForAll(J, x -> not IsBool(SolutionMat(I,x))) then return false; fi;

    if Length(K) = 0 then return true; fi;

    # check if L2 + [L2,L2] is an ideal
    if ForAll(J, x -> not IsBool(SolutionMat(K,x))) then return false; fi;
    
    return true;
end;

RelevantJNFs := function(n, F, flag)
    local jnfs;

    jnfs := JordanNormalForms(n,F,flag);
    jnfs := Filtered(jnfs, x -> x <> 0*x);
    jnfs := Filtered(jnfs, IsLieRelevant1);
    jnfs := Filtered(jnfs, IsLieRelevant2);
    jnfs := Filtered(jnfs, IsLieRelevant3);
    return jnfs; 

end;

