
# The classical simple Lie algebras over F up to dim 100
# See the book of Kac and Moody or the book by Willem

ClassicSimpleLieAlgebra := function( F, type, n )
    local L, I, c, p;

    p := Characteristic(F);

    if type = "A" then                            # A/Z(A) is simple
        if not n >= 1 then return fail; fi;
        L := SimpleLieAlgebra( "A", n, F );
        if IsInt((n+1)/p) then L := L/LieCentre(L); fi;
    fi;

    if type = "B" then                            # B/S(B) is simple
        if not n >= 2 then return fail; fi;
        L := SimpleLieAlgebra( "B", n, F  );
        if p = 2 then 
            c := ChiefSeriesLieAlgebra(L);
            L := L/c[Length(c)-1];
        fi;
    fi;

    if type = "C" then                            # C^(2)/Z(C^(2)) is simple
        if not n >= 3 then return fail; fi;
        L := SimpleLieAlgebra( "C", n, F );
        if p = 2 then 
            L := LieDerivedSeries(L)[3];
            if IsInt(n/2) then L := L/LieCentre(L); fi;
        fi;
    fi;

    if type = "D" then                            # D/Z(D) is simple
        if not n >= 4 then return fail; fi;
        L := SimpleLieAlgebra( "D", n, F );
        L := L/LieCentre(L);
    fi;

    if type = "E" then                            # E is simple
        if not n in [6,7,8] then return fail; fi;
        L := SimpleLieAlgebra( "E", n, F );
        if p=3 and n=6 then L := L/LieCentre(L); fi;
        if p=2 and n=7 then L := L/LieCentre(L); fi;
    fi;
       
    if type = "F" then                            # F decomposes for p=2
        if n <> 4 then return fail; fi;
        L := SimpleLieAlgebra( "F", n, F );
        if p = 2 then 
            I := IdealByGenerators( L, BasisVectors(Basis(L)){[1]} );
            L := L/I;
            L!.sub := I;
        fi;
    fi;
       
    if type = "G" then                            # G is simple
        if n <> 2 then return fail; fi;
        L := SimpleLieAlgebra( "G", n, F );
        if p = 3 then 
            L := L/ChiefSeriesLieAlgebra(L)[1];   # tested: factors are isom
        fi;
    fi;

    SetName( L, Concatenation( type, String(n) ) );
    return L;
end;

ClassicSimpleLieAlgebras := function( F, limit )
    local all, n, m1, m2, a, p, L, d;

    all := [];

    p := Characteristic(F);
    m1 := RootInt(limit+2)-1;               # 2n^2 + n <= limit
    m2 := QuoInt(RootInt( 2*limit+5 )+2,2); # 2n^2-n-2 <= limit

    # type A 
    if p=2 then a := 2; else a := 1; fi;
    for n in [a..m1] do 
        if IsInt((n+1)/p) then 
            d := n*(n+2)-1;
        else
            d := n*(n+2);
        fi;
        if d <= limit then 
            L := ClassicSimpleLieAlgebra(F, "A", n);
            if Dimension(L) <> d then Error("something wrong with A"); fi;
            Add(all, L);
        fi;
    od; 
Print("A done \n");

    # types B and C have dimension n(2n+1) (unless p=2)
    if p=2 then 
        for n in [3..m2] do
            if IsInt(n/2) then 
                d := 2*n^2-n-2;
            else
                d := 2*n^2-n-1;
            fi;
            if d <= limit then  
                L := ClassicSimpleLieAlgebra(F, "B", n);
                if Dimension(L) <> d then Error("something wrong with B"); fi;
                Add(all, L);
                L := ClassicSimpleLieAlgebra(F, "C", n);
                if Dimension(L) <> d then Error("something wrong with C"); fi;
                Add(all, L);
            fi;
        od;
    else
        for n in [2..m2] do
            d := n*(2*n+1);
            if d <= limit then  
                L := ClassicSimpleLieAlgebra(F, "B", n);
                if Dimension(L) <> d then Error("something wrong with B"); fi;
                Add(all, L);
            fi;
            if n > 2 and d <= limit then  
                L := ClassicSimpleLieAlgebra(F, "C", n);
                if Dimension(L) <> d then Error("something wrong with C"); fi;
                Add(all, L);
            fi;
        od;
    fi;
Print("B and C done \n");

    # type D has dimension n(2n-1)
    for n in [4..m2] do
        d := n*(2*n-1);
        if p=2 and IsInt(n/2) then d := d-2; elif p=2 then d:= d-1; fi;
        if d <= limit then 
            L := ClassicSimpleLieAlgebra(F, "D", n);
            if Dimension(L) <> d then Error("something wrong with D"); fi;
            Add(all, L);
        fi;
    od;
Print("D done \n");

    # type F
    if p=2 and limit >= 26 then 
        Add( all, ClassicSimpleLieAlgebra(F,"F",4)); 
    fi;
    if p<>2 and limit >= 52 then 
        Add( all, ClassicSimpleLieAlgebra(F,"F",4)); 
    fi;

    # type G
    if p<>3 and limit >=  14 then 
        Add( all, ClassicSimpleLieAlgebra(F,"G",2)); 
    fi;
    if p=3 and limit >=  7 then 
        Add( all, ClassicSimpleLieAlgebra(F,"G",2)); 
    fi;

    # type E
    if p<>3 and limit >= 78 then 
        Add(all, ClassicSimpleLieAlgebra(F,"E",6)); 
    fi;
    if p=3 and limit >= 77 then 
        Add(all, ClassicSimpleLieAlgebra(F,"E",6)); 
    fi;
    if p<>2 and limit >= 133 then 
       Add( all, ClassicSimpleLieAlgebra(F,"E",7)); 
    fi;
    if p=2 and limit >= 132 then 
       Add( all, ClassicSimpleLieAlgebra(F,"E",7)); 
    fi;
    if limit >= 248 then Add( all, ClassicSimpleLieAlgebra(F,"E",8)); fi;

    Sort( all, function(a,b) return Dimension(a)<Dimension(b); end );
    return all;
end;


